/**
 * File:
 *  modbusSlave.c
 *
 * Notes:
 *  This file contains the implementation for a serial RTU modbus slave.
 *
 * Functions:
 *  addCoil         Add a coil to the slave
 *  addHoldingReg   Add a holding register to the slave
 *  addInputReg     Add an input register to the slave
 *  addStatusBit    Add a status bit to the slave
 *  calcCRC         Calculates the CRC for the message content
 *  coilState       Sets the state of a coil
 *  decodePacket    Decodes a packet received from modbus master
 *  modbusSlave     Initialise modbus slave structure
 *  packBits        Packs bits into a message buffer
 *  packRegisters   Packs registers into a message buffer
 *  setRegister     Sets the value of a holding register
 *  startTimeout    Starts the message timeout timer
 *  restartRx       Restart communications
 *  validAddresses  Validates an address
 *  validDiscretes  Validates a discrete
 *  validRegisters  Validates a register
 *
 * History:
 *  10/06/2012 Written by Simon Platten
 *  19/06/2012 Added receivePacket function which buffers a packet
 *  21/06/2012 Implemented exception responses
 *             Implemented Timer0 to time receipt of bytes
 *  25/06/2012 Implemented startTimeout
 *  27/06/2012 Replaced individual request and response buffers with a single
 *             message buffer.  Re-wrote CRC to use lookup tables.  Improved
 *             comments.
 *  30/06/2012 Implemented register mappings to allow 8 bit I/O ports to be
 *             mapped into 16 bit registers
 */
#include <built_in.h>

#include "modbusSlave.h"
// Pointers to last node in each linked list
static coilDef*    pLastCoil       = NULL;
static statusDef*  pLastStatusBit  = NULL;
static inputDef*   pLastInput      = NULL;
static holdingDef* pLastHoldingReg = NULL;
// Pointers to linked lists for each data type
static coilDef*    pCoils          = NULL;
static statusDef*  pStatusBits     = NULL;
static inputDef*   pInputs         = NULL;
static holdingDef* pHoldingRegs    = NULL;
// Flags to use when validating address ranges
static boolean blnValidStart, blnValidEnd;
// Exception code
static mbException exceptionCode;
// The modbus slave address
static byte bytSlaveAddress;
// The requested start address high byte and low byte
static byte bytStartHi, bytStartLo;
// The requested total number of items high byte and low byte
static byte bytTotalHi, bytTotalLo;
// The byte index
static byte bytIndex = 0;
// Receiver & transmitter GAP set-points
static uint uintRxGapSetPt1;
static uint uintRxGapSetPt2;
static uint uintRxGapSetPt3;
static uint uintRxGapSetPt;
// Receiver GAP counter
static uint uintRxGap;
// Received packet buffer
static byte aryBuffer[MAX_PACKET_LENGTH];
static byte bytRxphase;
// RS-485 direction control bit
sbit Tx_dir at LATE.B0;
/**
 * Function:
 *  addDiscrete
 *
 * Parameters:
 *  pDiscrete, pointer to an instance of discreteDef
 *  uintAddress, the address to assign to the discrete (base 1)
 *  pByte, pointer to the byte containing the bit to map as the discrete
 *  bBit, the bit in the byte to map to the discrete, 0 to 7
 *  pLinkList, pointer to pointer of discrete linked list
 *  pLast, pointer to pointer of last node in list
 *
 * Returns:
 *  TRUE if discrete added, FALSE if not
 */
static boolean addDiscrete(discreteDef* pDiscrete,
                           uint uintAddress,
                           ushort* pByte, byte bBit,
                           discreteDef** pLinkList,
                           discreteDef** pLast) {
  if ( !(pDiscrete && uintAddress && pByte &&
         bBit >= 0 && bBit < 8 && pLinkList && pLast) ) {
    return FALSE;
  }
  if ( *pLinkList == NULL ) {
// This is the first coil
    *pLinkList = *pLast = pDiscrete;
  } else {
// Append this coil to the linked list
    (*pLast)->pNext = pDiscrete;
// Adjust the last node
    *pLast = pDiscrete;
  }
// Ensure the discrete is set-up using the passed parameters
  pDiscrete->uintAddress = uintAddress;
  pDiscrete->pByte = pByte;
  pDiscrete->bBit  = bBit;
  pDiscrete->pNext = NULL;
}
/**
 * Function:
 *  addRegister
 *
 * Parameters:
 *  pRegister, pointer to an instance of registerDef
 *  uintAddress, the address to assign to the register (base 1)
 *  bytMapping, 0 = map pWord to low byte of register
 *              1 = map pWord to high byte of register
 *  pWord, pointer to the word to map as the register
 *  pLinkList, pointer to pointer of register linked list
 *  pLast, pointer to pointer of last node in list
 *  pWord, pointer to the word to use for the holding register
 *
 * Notes:
 *  The Mapping parameter allows 8 bit registers to be mapped into a 16 bit
 * register.  To map a normal 16 bit register, bytMapping should be 0
 *
 * Returns:
 *  TRUE if register added, FALSE if not
 */
static boolean addRegister(registerDef* pRegister,
                           uint uintAddress,
                           byte bytMapping,
                           uint* pWord,
                           registerDef** pLinkList,
                           registerDef** pLast) {
  if ( !(pRegister && uintAddress && pWord && pLinkList && pLast &&
         (bytMapping == 0 || bytMapping == 1)) ) {
    return FALSE;
  }
// Is this register already set-up ?
  if ( pRegister->uintAddress == uintAddress ) {
// Yes, is there a matching mapping ?
    if ( (bytMapping == 0 && pRegister->mapping.pBase != NULL) ||
         (bytMapping == 1 && pRegister->mapping.pLoHi[1] != NULL) ) {
// Yes, abort!
      return FALSE;
    }
  } else {
// This register doesn't exist, add it to the linked list now
    if ( *pLinkList == NULL ) {
// This is the first register
      *pLinkList = *pLast = pRegister;
    } else {
// Append this register to the linked list
      (*pLast)->pNext = pRegister;
// Adjust the last node
      *pLast = pRegister;
    }
// Ensure the discrete is set-up using the passed parameters
    pRegister->uintAddress = uintAddress;
    pRegister->pNext = NULL;
  }
  if ( bytMapping == 0 ) {
    pRegister->mapping.pBase = pWord;
  } else {
    pRegister->mapping.pLoHi[1] = (byte*)pWord;
  }
}
/**
 * Function:
 *  calcCRC
 *
 * Parameters:
 *  uintLength, the number of bytes in the message
 *
 * Returns:
 *  A 16bit CRC for the passed message
 */
static long calcCRC(uint uintLength) {
  long lngCRC = 0xffff;
  uint i, j;
  for( i=0; i<uintLength; i++ ) {
    lngCRC ^= (aryBuffer[i] & 0xff);

    for( j=0; j<8; j++ ) {
      if ( lngCRC & 0x01 ) {
        lngCRC >>= 0x01;
        lngCRC ^= 0xa001;
      } else {
        lngCRC >>= 0x01;
      }
    }
  }
  return lngCRC;
}
/**
 * Function:
 *  coilState
 *
 * Parameters:
 *  uintAddr, the address of the coil to modify
 *  eState, TRUE to turn coil on, FALSE to turn coil off
 *
 * Returns:
 *  none
 */
static void coilState(uint uintAddr, boolean eState) {
// Find the first I/O point
  discreteDef* pNode;
  byte bytMask;
  for( pNode=pCoils; pNode!=NULL; pNode=pNode->pNext ) {
    if ( pNode->uintAddress != uintAddr  ) {
      continue;
    }
    bytMask = (1 << pNode->bBit);

    if ( eState == TRUE ) {
// Turn on
      *pNode->pByte |= bytMask;
    } else {
// Turn off
      *pNode->pByte &= (0xffff ^ bytMask);
    }
    break;
  }
}
/**
 * Function:
 *  packBits
 *
 * Parameters:
 *  pLL, a pointer to the linked list containing the discrete items
 *  pBuffer, a pointer to the buffer to pack the bits into
 *  uintStart, the start address
 *  uintEnd, the end address
 *
 * Returns:
 *  The number of bytes the bits were packed into
 */
static uint packBits(discreteDef* pLL, byte* pBuffer, 
                     uint uintStart, uint uintEnd) {
// Find the first I/O point
  uint uintBytes = 0, uintBit = 0;
  discreteDef* pNode;
  byte bytSet;
  for( pNode=pLL; pNode!=NULL; pNode=pNode->pNext ) {
    if ( pNode->uintAddress >= uintStart && pNode->uintAddress <= uintEnd ) {
      bytSet = (*pNode->pByte & (1 << pNode->bBit)) ? 1 : 0;
      
      if ( bytSet ) {
        pBuffer[uintBytes] |= (bytSet << uintBit);
      }
      if ( ++uintBit > 7 ) {
        uintBit = 0;
        uintBytes++;
      }
    }
  }
  return uintBytes + 1;
}
/**
 * Function:
 *  packRegisters
 *
 * Parameters:
 *  pLL, a pointer to the linked list containing the register items
 *  pBuffer, a pointer to the buffer to pack the bits into
 *  uintStart, the start address
 *  uintEnd, the end address
 *
 * Returns:
 *  The number of bytes the registers were packed into
 */
static uint packRegisters(registerDef* pLL, byte* pBuffer,
                          uint uintStart, uint uintEnd) {
// Find the first I/O point
  uint uintBytes = 0;
  registerDef* pNode;
  for( pNode=pLL; pNode!=NULL; pNode=pNode->pNext ) {
    if ( pNode->uintAddress >= uintStart && pNode->uintAddress <= uintEnd ) {
      if ( pNode->mapping.pLoHi[1] != NULL ) {
        pBuffer[uintBytes++] = *pNode->mapping.pLoHi[1];
        pBuffer[uintBytes++] = *pNode->mapping.pLoHi[0];
      } else {
        pBuffer[uintBytes++] = (*pNode->mapping.pBase >> 0x08);
        pBuffer[uintBytes++] = (*pNode->mapping.pBase & 0xff);
      }
    }
  }
  return uintBytes;
}
/**
 * Function:
 *  setRegister
 *
 * Parameters:
 *  uintAddr, the address of the register to modify
 *  pData, a pointer to the value
 *
 * Returns:
 *  none
 */
static void setRegister(uint uintAddr, uint* pData) {
// Find the first I/O point
  registerDef* pNode;
  uint uintTemp;
  uintTemp = *pData;
  for( pNode=pHoldingRegs; pNode!=NULL; pNode=pNode->pNext ) {
    if ( pNode->uintAddress != uintAddr  ) {
      continue;
    }
    if ( pNode->mapping.pLoHi[1] != NULL ) {
     *pNode->mapping.pLoHi[0] = Hi(uintTemp);
     *pNode->mapping.pLoHi[1] = Lo(uintTemp);
    } else {
      *pNode->mapping.pBase = uintTemp;
    }
    break;
  }
}
/**
 * Function:
 *  startTimeout
 *
 * Parameters:
 *  none
 *
 * Returns:
 *  none
 */
static void startTimeout(void) {
// Disable interrupt
  TMR0ON_bit = 0;
// Set timer presets
  if ( bytRxphase.B1 ) {
    TMR0H = Hi(uintRxGapSetPt3);
    TMR0L = Lo(uintRxGapSetPt3);
  } else if ( bytRxphase.B2 ) {
    TMR0H = Hi(uintRxGapSetPt2);
    TMR0L = Lo(uintRxGapSetPt2);
  } else if ( bytRxphase.B7 ) {
    TMR0H = Hi(uintRxGapSetPt1);
    TMR0L = Lo(uintRxGapSetPt1);
  } else {
    TMR0H = Hi(uintRxGapSetPt);
    TMR0L = Lo(uintRxGapSetPt);
  }
// Clear the interrupt mask
  TMR0IF_bit = 0;
// Enable interrupt
  TMR0ON_bit = 1;
}
/**
 * Function:
 *  restartRx
 *
 * Parameters:
 *  none
 *
 * Returns:
 *  none
 */
static void restartRx(void) {
// Clear FIFO
  bytRxphase = RCREG;
  bytRxphase = RCREG;
// Clear errors and enable Rx
  CREN_bit = 0;
  CREN_bit = 1;
// Set RS-485 direction
  Tx_dir = 0;
// Set parity check
  //RX9_bit = 1;
// ...or double STOP bit check
  //RX9D_bit = 1;
// Clear counters
  bytRxphase = 0;
  bytIndex = 0;
// Disable Tx
  TXEN_bit = 0;
  TXIE_bit = 0;
// Enable Rx interrupts
  RCIE_bit = 1;
  startTimeout();
}
/**
 * Function:
 *  validAddresses
 *
 * Parameters:
 *  uintAddress, the address to validate against
 *  uintStart, the first address to validate
 *  uintEnd, the last address to validate
 *
 * Returns:
 *  TRUE of FALSE if both addresses are valid
 */
static boolean validAddresses(uint uintAddress, uint uintStart, uint uintEnd) {
  if ( blnValidStart == FALSE && uintStart == uintAddress ) {
    blnValidStart = TRUE;
  }
  if ( blnValidEnd == FALSE && uintEnd == uintAddress ) {
    blnValidEnd = TRUE;
  }
  return (blnValidStart == TRUE && blnValidEnd == TRUE);
}
/**
 * Function:
 *  validDiscretes
 *
 * Parameters:
 *  pLL, a pointer to the linked list containing the discrete items
 *  uintStart, the start address
 *  uintEnd, the end address
 *
 * Returns:
 *  TRUE if the start and end addresses are valid, FALSE if not
 */
static boolean validDiscretes(discreteDef* pLL, uint uintStart, uint uintEnd) {
  discreteDef* pNode;
  blnValidStart = blnValidEnd = FALSE;
  for( pNode=pLL; pNode!=NULL; pNode=pNode->pNext ) {
    if ( validAddresses(pNode->uintAddress, uintStart, uintEnd) == TRUE ) {
      return TRUE;
    }
  }
  return FALSE;
}
/**
 * Function:
 *  validRegisters
 *
 * Parameters:
 *  pLL, a pointer to the linked list containing the register items
 *  uintStart, the start address
 *  uintEnd, the end address
 *
 * Returns:
 *  TRUE if the start and end addresses are valid, FALSE if not
 */
static boolean validRegisters(registerDef* pLL, uint uintStart, uint uintEnd) {
  registerDef* pNode;
  blnValidStart = blnValidEnd = FALSE;
  for( pNode=pLL; pNode!=NULL; pNode=pNode->pNext ) {
    if ( validAddresses(pNode->uintAddress, uintStart, uintEnd) == TRUE ) {
      return TRUE;
    }
  }
  return FALSE;
}
/**
 * Function:
 *  addCoil
 *
 * Parameters:
 *  pCoil, pointer to an instance of coilDef
 *  uintAddress, the address to assign to the coil (base 1)
 *  pByte, pointer to the byte containing the bit to map as the coil
 *  bBit, the bit in the byte to map to the coil, 0 to 7
 *
 * Returns:
 *  TRUE if coil added, FALSE if not
 */
boolean addCoil(coilDef* pCoil, uint uintAddress, ushort* pByte, byte bBit) {
  return addDiscrete(pCoil, uintAddress, pByte, bBit, &pCoils, &pLastCoil);
}
/**
 * Function:
 *  addHoldingReg
 *
 * Parameters:
 *  pHolding, pointer to an instance of holdingDef
 *
 * Notes:
 *  see addRegister for further parameter descriptions
 *
 * Returns:
 *  TRUE if coil added, FALSE if not
 */
boolean addHoldingReg(holdingDef* pHolding, uint uintAddress, byte bytMapping,
                      uint* pWord) {
  return addRegister(pHolding, uintAddress, bytMapping, pWord, &pHoldingRegs,
                     &pLastHoldingReg);
}
/**
 * Function:
 *  addInputReg
 *
 * Parameters:
 *  pInput, pointer to an instance of inputDef
 *
 * Notes:
 *  see addRegister for further parameter descriptions
 *
 * Returns:
 *  TRUE if coil added, FALSE if not
 */
boolean addInputReg(inputDef* pInput, uint uintAddress, byte bytMapping,
                    uint* pWord) {
  return addRegister(pInput, uintAddress, bytMapping, pWord, &pInputs,
                     &pLastInput);
}
/**
 * Function:
 *  addStatusBit
 *
 * Parameters:
 *  pStatus, pointer to an instance of statusDef
 *  uintAddress, the address to assign to the status bit (base 1)
 *  pByte, pointer to the byte containing the bit to map as the status bit
 *  bBit, the bit in the byte to map to the status bit, 0 to 7
 *
 * Returns:
 *  TRUE if status bit added, FALSE if not
 */
boolean addStatusBit(statusDef* pStatus, uint uintAddress, ushort* pByte, 
                     byte bBit) {
  return addDiscrete(pStatus, uintAddress, pByte, bBit, &pStatusBits, 
                     &pLastStatusBit);
}
/**
 * Function:
 *  decodePacket
 *
 * Parameters:
 *  none
 *
 * Returns:
 *  none
 */
void decodePacket(void) {
  uint uintCRCpos, uintCRC;

  if ( TMR0IF_bit ) {
    bytRxphase.B0 = 1;
    TMR0ON_bit = 0;
// Clear the interrupt mask
    TMR0IF_bit = 0;

    if ( bytRxphase.B7 ) {
// Last byte sent, restart reception & set silent interval
      restartRx();
      return;
    }
  }
  if( RCIF_bit ) {
    byte bytTemp;
// Overrun or framing error?
    if ( OERR_bit || FERR_bit ) {
      restartRx();
      return;
    }
// May need to store RX9D_bit here for parity check!
// Read received byte
    bytTemp = RCREG;
// Silent interval or inter-char interval passed?
// If passed, is there enough space in buffer?
    if ( bytRxphase == 0 || bytRxphase.B2 == 1 ||
         bytIndex >= MAX_PACKET_LENGTH ) {
// No, wait for end of current message
      bytIndex = 0;
      bytRxphase = 0;
      startTimeout();  // 4.5 chars
      return;
    }
    if ( bytIndex == 0) {
// Address received
      bytRxphase = 2;
    }
// Store the data in buffer
    aryBuffer[bytIndex++] = bytTemp;
// Set inter-char interval
    startTimeout();  // 2.5 chars
  } else if ( bytRxphase.B7 ) {
    if ( TXIF_bit ) {
      if ( bytIndex < aryBuffer[0] ) {
        TXREG=aryBuffer[bytIndex++];
      } else {
// Wait for last byte to be transmitted
        startTimeout();  // 1 char
// Disable Tx interrupts
        TXIE_bit = 0;
      }
      return;
    }
  }
  if ( bytRxphase == 3 ) {
// Inter-char interval passed, check inter-frame timeout remainder
    bytRxphase = 4;
    startTimeout(); // 2 chars
  }
  if ( bytRxphase != 5 ) {
// Nothing more to do
    return;
  }
// Message received, our addreess?
  if (aryBuffer[0] == bytSlaveAddress ) {
// Set the expected position of the CRC
    uintCRCpos = 6;
// Is the function supported?
    switch( aryBuffer[1] ) {
    case READ_COILS:
    case FORCE_SINGLE_COIL:
    case FORCE_MULTIPLE_COILS:
      if ( pCoils == NULL ) {
// No coils defined!
        exceptionCode = ILLEGAL_DATA_ADDRESS;
      }
      break;
    case READ_STATUS_INPUTS:
      if ( pStatusBits == NULL ) {
// No status inputs defined!
        exceptionCode = ILLEGAL_DATA_ADDRESS;
      }
      break;
    case READ_HOLDING_REGISTERS:
    case PRESET_SINGLE_REGISTER:
    case PRESET_MULTIPLE_REGISTERS:
      if ( pHoldingRegs == NULL ) {
// No holding registers defined!
        exceptionCode = ILLEGAL_DATA_ADDRESS;
      }
      break;
    case READ_INPUT_REGISTERS:
      if ( pInputs == NULL ) {
// No input registers defined!
        exceptionCode = ILLEGAL_DATA_ADDRESS;
      }
      break;
    default:
// No, function not supported!
      exceptionCode = ILLEGAL_FUNCTION;
      break;
    }
    bytStartHi = aryBuffer[2];
    bytStartLo = aryBuffer[3];
    bytTotalHi = aryBuffer[4];
    bytTotalLo = aryBuffer[5];

    if ( aryBuffer[1] == FORCE_MULTIPLE_COILS ||
         aryBuffer[1] == PRESET_MULTIPLE_REGISTERS )  {
      uintCRCpos += 1 + aryBuffer[6];
    }
// Calculate the packet CRC
    uintCRC = calcCRC(uintCRCpos);
// Compare the calculate checksum with the received checksum
    if ( (uintCRC & 0xff)  == aryBuffer[uintCRCpos] &&
         (uintCRC >> 0x08) == aryBuffer[uintCRCpos + 1] ) {
// CRC ok
      uint uintSAddr, uintEAddr, uintItemCount, uintLength = 6;
      uint uintOffset, uintAddr, uintBit;
      discreteDef* pDiscretes;
      registerDef* pRegisters;
// Get the start address
      uintSAddr = (((uint)bytStartHi << 0x08) | bytStartLo) + 1;
// How many items have been requested?
      uintItemCount = ((uint)bytTotalHi << 0x08) | bytTotalLo;
// Whats the last address?
      uintEAddr = uintSAddr + uintItemCount - 1;
// What was the requested function code?
      switch( aryBuffer[1] ) {
      case READ_COILS:
      case READ_STATUS_INPUTS:
        if ( aryBuffer[1] == READ_COILS ) {
          pDiscretes = pCoils;
        } else {
          pDiscretes = pStatusBits;
        }
        if ( (uintEAddr - uintSAddr + 1) > MAX_DISCRETES_IN_1_AND_2 ||
             validDiscretes(pDiscretes, uintSAddr, uintEAddr) == FALSE ) {
          exceptionCode = ILLEGAL_DATA_ADDRESS;
// Exception, address does not exist
        } else {
          aryBuffer[2] = packBits(pDiscretes,
                                  &aryBuffer[3],
                                  uintSAddr, uintEAddr);
          uintLength = aryBuffer[2] + 3;
        }
        break;
      case READ_HOLDING_REGISTERS:
      case READ_INPUT_REGISTERS:
        if ( aryBuffer[1] == READ_INPUT_REGISTERS ) {
          pRegisters = pInputs;
        } else {
          pRegisters = pHoldingRegs;
        }
        if ( (uintEAddr - uintSAddr + 1) > MAX_REGISTERS_IN_3_AND_4 ||
             validRegisters(pRegisters, uintSAddr, uintEAddr) == FALSE ) {
// Exception, address does not exist
          exceptionCode = ILLEGAL_DATA_ADDRESS;
        } else {
          aryBuffer[2] = packRegisters(pRegisters,
                                       &aryBuffer[3],
                                       uintSAddr, uintEAddr);
          uintLength = aryBuffer[2] + 3;
        }
        break;
      case FORCE_SINGLE_COIL:
        if ( validDiscretes(pCoils, uintSAddr, uintSAddr) == FALSE ) {
// Exception, address does not exist
          exceptionCode = ILLEGAL_DATA_ADDRESS;
        } else {
          if ( aryBuffer[4] == 0xff && aryBuffer[5] == 0x0 ) {
// Force coil on
            coilState(uintSAddr, TRUE);
          } else if ( aryBuffer[4] == 0x0 && aryBuffer[5] == 0x0 ) {
// Force coil off
            coilState(uintSAddr, FALSE);
          }
        }
        break;
      case PRESET_SINGLE_REGISTER:
        if ( validRegisters(pHoldingRegs, uintSAddr, uintSAddr) == FALSE ) {
// Exception, address does not exist
          exceptionCode = ILLEGAL_DATA_ADDRESS;
        } else {
          setRegister(uintSAddr, (uint*)&aryBuffer[4]);
        }
        break;
      case FORCE_MULTIPLE_COILS:
        if ( uintItemCount > MAX_DISCRETES_IN_FC15 ||
             uintItemCount > (aryBuffer[6] * 8) ) {
// Exception, either to many items or the coil count doesn't match byte count
          exceptionCode = ILLEGAL_DATA_ADDRESS;
          break;
        }
      case PRESET_MULTIPLE_REGISTERS:
        if ( (aryBuffer[1] == PRESET_MULTIPLE_REGISTERS &&
             uintItemCount > MAX_REGISTERS_IN_FC16) ) {
// Exception, either to many items or the coil count doesn't match byte count
          exceptionCode = ILLEGAL_DATA_ADDRESS;
        } else if ( validDiscretes(pCoils,
                                   uintSAddr, uintEAddr) == FALSE ) {
          aryBuffer[1] |= EXCEPTION_FLAG;
        } else {
          uintOffset = 0;
          uintAddr = uintSAddr;

          while( uintAddr<=uintEAddr ) {
            if ( aryBuffer[1] == PRESET_MULTIPLE_REGISTERS ) {
              setRegister(uintAddr,
                          (uint*)&aryBuffer[7 + uintOffset]);
              uintOffset += 2;
            } else {
              for(uintBit=0; uintBit<8; uintBit++, uintAddr++ ) {
                coilState(uintAddr,
                          aryBuffer[7 + uintOffset] & (1 << uintBit));
              }
              uintOffset++;
            }
            uintAddr++;
          }
        }
        break;
      }
      if( exceptionCode != NO_EXCEPTION ) {
        aryBuffer[1] |= EXCEPTION_FLAG;
        aryBuffer[2] = (byte)exceptionCode;
        exceptionCode = NO_EXCEPTION;
        uintLength = 3;
      }
      if ( uintLength > 0 ) {
// Calculate CRC for response
        uintCRC = calcCRC(uintLength);
        aryBuffer[uintLength] = (uintCRC & 0xff);
        uintLength++;
        aryBuffer[uintLength] = (uintCRC >> 0x08);
        uintLength++;
// Stop receiving
        CREN_bit = 0;
        RCIE_bit = 0;
// Enable transmission, set direction
        TXEN_bit = 1;
        Tx_dir = 1;
// Mark start of transmission
        bytRxphase = 128;
// Send address
        TXREG = bytSlaveAddress;
// Store length and set index
        aryBuffer[0] = uintLength;
        bytIndex = 2;
// Send function code to fill Tx buffer and avoid instant return to ISR
        while( !TXIF_bit ){}
        TXREG = aryBuffer[1];
// Enable Tx interrupts for remainder of response to follow
        TXIE_bit = 1;
        return;
      }
    }
  }
// No need to answer (not our address or message frame error, await next one)
  bytIndex = 0;
  bytRxphase = 1;
}
/**
 * Function:
 * modbusSlave
 *
 * Parameters:
 *  eBaud, see modbusSlave.h for options
 *  bytAddress, the modbus slave address 1 to 255
 *  bytStopBits, number of transmitted stop bits, 1 or 2
 *
 * Returns:
 *  0 if ok, -1 if error
 */
int modbusSlave(baudRate eBaud, const byte bytAddress, const byte bytStopBits) {
  long lngTimeoutPreset;
  byte bytTemp;

  if ( bytAddress < 1 || bytAddress > 247 ) {
    return -1;
  }
// Make sure the received array is cleared
  memset(aryBuffer, 0, sizeof(aryBuffer));
// Set the slave address
  bytSlaveAddress = bytAddress;
// Initialise the Rx Gap counter
  uintRxGap = 0;
// Ensure there is no exception code
  exceptionCode = NO_EXCEPTION;
// Initialize UART module for the specified bps
  switch( eBaud ) {
  case BAUD_1200:
    UART1_Init(BAUD_1200 * 100);
    lngTimeoutPreset = GAP_SETPT_1200;
    break;
  case BAUD_2400:
    UART1_Init(BAUD_2400 * 100);
    lngTimeoutPreset = GAP_SETPT_2400;
    break;
  case BAUD_4800:
    UART1_Init(BAUD_4800 * 100);
    lngTimeoutPreset = GAP_SETPT_4800;
    break;
  case BAUD_9600:
    UART1_Init(BAUD_9600 * 100);
    lngTimeoutPreset = GAP_SETPT_9600;
    break;
  case BAUD_19200:
    UART1_Init(BAUD_19200 * 100);
    lngTimeoutPreset = GAP_SETPT_19200;
    break;
 case BAUD_38400:
    UART1_Init(BAUD_38400 * 100);
    lngTimeoutPreset = GAP_SETPT_38400;
    break;
  case BAUD_57600:
    UART1_Init(BAUD_57600 * 100);
    lngTimeoutPreset = GAP_SETPT_57600;
    break;
  case BAUD_115200:
    UART1_Init(BAUD_115200 * 100);
    lngTimeoutPreset = GAP_SETPT_115200;
    break;
  default:
    return -1;
  }
// Wait for UART module to stabilize
  Delay_ms(100);
// Timer0 Registers:
// 16-Bit Mode
// Prescaler=1:1
  T0CON.TMR0ON = 0;// Timer0 On/Off Control bit 1=Enables / 0=Stops
  T0CON.T08BIT = 0;// Timer0 8-bit/16-bit Control bit: 1=8-bit/ 0=16-bit
  T0CON.T0CS   = 0;// TMR0 Clock Source Select bit: 0=Internal
  T0CON.T0SE   = 0;// TMR0 Source Edge Select bit: 0=low/high
  T0CON.PSA    = 1;// Prescaler Assignment bit: 1=NOT assigned/bypassed
  T0CON.T0PS2  = 0;// bits 2-0  PS2:PS0: Prescaler Select bits
  T0CON.T0PS1  = 0;
  T0CON.T0PS0  = 0;
// Set prescaler if needed
  bytTemp = 0;
  while( lngTimeoutPreset > 29126 ) {
    lngTimeoutPreset >>= 1;
    bytTemp++;
  }
  if ( bytTemp > 0 && bytTemp <= 8 ) {
    T0CON.PSA = 0;   // Prescaler assigned
    T0CON = T0CON | (--bytTemp & 0x07);
  }
  uintRxGapSetPt1 = (uint)-(lngTimeoutPreset >> 1);   // 1 char
  uintRxGapSetPt2 = (uint)-lngTimeoutPreset;          // 2 chars
  uintRxGapSetPt3 = (uint)-(lngTimeoutPreset*5 >> 2); // 2.5 chars
  uintRxGapSetPt  = (uint)-(lngTimeoutPreset*9 >> 2); // 4.5 chars
// Double STOP bit while transmitting?
  if ( bytStopBits != 1 ) {
    TX9_bit=1;
    TX9D_bit=1;
  }
// Clear errors, Rx buffer & Rx phase indicator
// and start silent interval in case bus is active
  restartRx();
  TMR0IE_bit = 1;
// Enable peripheral interrupts
  PEIE_bit   = 1;
// Enable GLOBAL interrupts
  GIE_bit    = 1;
}